//
//  GLViewController.m
//  UsingTheTemplate
//
//  Created by Yuji Nomura on 14/05/01.
//

#import "GLViewController.h"
#import "ConstantsAndMacros.h"
#import "OpenGLCommon.h"
#import "Ground.h"
#import "P2Object.h"
#include "MyQuat.h"

//----------------------------------------------------------------------------------------------------------
int W = 300, H = 300;
int clicked_x, clicked_y;
Quat4f current_rot = UNIT_QUAT4F_CONSTANT;  /* no rotation */
Quat4f dragged_rot = UNIT_QUAT4F_CONSTANT;
Quat4f spring_rot = UNIT_QUAT4F_CONSTANT;

/* Note: a Lx + b Ly + c Lz > 0 must be satisfied (normal (a,b,c) directs to light side */
GLfloat *getShadowMatrix(
                         GLfloat a, GLfloat b, GLfloat c, GLfloat d,  /* of ax+by+cz+d=0 plane */
                         GLfloat Lx, GLfloat Ly, GLfloat Lz,  /* light position of Lw=1 */
                         GLfloat *m  /* where to output */
                         ){
    GLfloat Lw = 1.0f;
    GLfloat ip; /* inner product */
    
    ip = a * Lx + b * Ly + c * Lz + d * Lw;  /* N^T L */
    
    /* projection matrix = N^T L I - L N^T */
    m[0] = -a*Lx; m[4] = -b*Lx;  m[8] = -c*Lx; m[12] = -d*Lx;
    m[1] = -a*Ly; m[5] = -b*Ly;  m[9] = -c*Ly; m[13] = -d*Ly;
    m[2] = -a*Lz; m[6] = -b*Lz; m[10] = -c*Lz; m[14] = -d*Lz;
    m[3] = -a*Lw; m[7] = -b*Lw; m[11] = -c*Lw; m[15] = -d*Lw;
    
    m[0] += ip;
    m[5] += ip;
    m[10] += ip;
    m[15] += ip;
    
    /* 	printf("%.3f %.3f %.3f %.3f \n" */
    /* 		   "%.3f %.3f %.3f %.3f \n" */
    /* 		   "%.3f %.3f %.3f %.3f \n" */
    /* 		   "%.3f %.3f %.3f %.3f \n", */
    /* 		   m[ 0], m[ 1], m[ 2], m[ 3], */
    /* 		   m[ 4], m[ 5], m[ 6], m[ 7], */
    /* 		   m[ 8], m[ 9], m[10], m[11], */
    /* 		   m[12], m[13], m[14], m[15]  ); */
    
    return m;
}

#define ROTATE_GROUND
#define CLIP_SHADOW_TO_GROUND
void draw()
{
	const GLfloat lightpos[] = {3.0, 4.0, 5.0, 1.0};
	/*   GLfloat shadowDiffuse[4] = {0.0f, 0.0f, 0.0f, 0.5f}; */
	GLfloat shadowDiffuse[4] = {0.0f, 0.0f, 0.0f, 1.0f};
	Quat4f tmp;
	float rotMatrix[16];
	GLfloat shadowMatrix[16];
    
    //    GLfloat ambient_bak[4], diffuse_bak[4];
	GLfloat a, b, c, d; /* of ax+by+cz+d=0 plane */
	
	//	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	//	glClear(GL_STENCIL_BUFFER_BIT);
    
	glLoadIdentity();  /* reset model view matrix */
    
	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_POSITION, lightpos); /* affected by model view matrix */
    
    //test
    //    glDisable(GL_TEXTURE_2D);
    //    glEnable(GL_COLOR_MATERIAL);
    //    glColor4f(0,0,0,0);
    //    glDisable(GL_COLOR_MATERIAL);
    
#ifdef CLIP_SHADOW_TO_GROUND
	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_ALWAYS, 1, 1);  /* make stencil value = 1 where ground is drawn */
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
#endif
    
#ifndef ROTATE_GROUND
	/* draw ground(not rotated) */
	[Ground draw];
#endif
    
	/* calculate mouse rotation matrix */
	mul_quat(&current_rot, &dragged_rot, &tmp);
	quat_to_matrix(&tmp, rotMatrix);
	glLoadMatrixf(rotMatrix);
    
#ifdef ROTATE_GROUND
	/* draw ground(rotated) */
	[Ground draw];
#endif
    
    //test
    //    glDisable(GL_TEXTURE_2D);
    //    glEnable(GL_COLOR_MATERIAL);
    //    glColor4f(0,0,0,0);
    //    glDisable(GL_COLOR_MATERIAL);
    
#ifdef CLIP_SHADOW_TO_GROUND
	glDisable(GL_STENCIL_TEST);
#endif
    
	/* draw objects */
    
	//	NSLog(@"before drawScene(): glGetError()=%d", glGetError());
	[P2Object draw];
	//	NSLog(@"after drawScene(): glGetError()=%d", glGetError());
	
	//	glDisable(GL_COLOR_MATERIAL);
	//	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient_bak);
	//	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse_bak);
	
	
	//tmp
    //	return;
	
	
	/* draw shadow */
	glDepthMask(GL_FALSE);  /* stop updating depth buffer */
    
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
	glPolygonOffset(-1, -1);  /* offset shadow's depth to a bit nearer */
	glEnable(GL_POLYGON_OFFSET_FILL);
    
	/* ground plane parameters */
#ifndef ROTATE_GROUND
	/* y+0.9=0 */
	a = 0;
	b = 1;
	c = 0;
	d = 0.9f;
#else
	/* rotate normal (0,1,0) of y+0.9=0 */
	a = rotMatrix[0] * 0 + rotMatrix[4] * 1 + rotMatrix[8] * 0;
	b = rotMatrix[1] * 0 + rotMatrix[5] * 1 + rotMatrix[9] * 0;
	c = rotMatrix[2] * 0 + rotMatrix[6] * 1 + rotMatrix[10] * 0;
	d = 0.9f;
#endif
	glLoadMatrixf(getShadowMatrix(a, b, c, d, lightpos[0], lightpos[1], lightpos[2], shadowMatrix));
	glMultMatrixf(rotMatrix);
    
	glEnable(GL_STENCIL_TEST);
#ifdef CLIP_SHADOW_TO_GROUND
	/* draw only where ground is drawn and shadow is not drawn */
	glStencilFunc(GL_EQUAL, 1, ~0);   /* draw only if stencil value = 1 (ground drawn) */
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);  /* stencil value = 2 if shadow drawn */
#else
	glClear(GL_STENCIL_BUFFER_BIT);
	/* in order to avoid shadow overlapping, draw then fill stencil */
	glStencilFunc(GL_NOTEQUAL, 1, 1);  /* draw only if stencil value = 0 */
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);  /* stencil value = 1 if drawn */
#endif
	
	glMaterialfv(GL_FRONT, GL_DIFFUSE, shadowDiffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, shadowDiffuse);
	//	NSLog(@"before drawScene(for shadow): glGetError()=%d", glGetError());
    //	[P2Object drawShadow];
	//	NSLog(@"after drawScene(for shadow): glGetError()=%d", glGetError());
	
	glDisable(GL_STENCIL_TEST);
	glDepthMask(GL_TRUE);
	glDisable(GL_POLYGON_OFFSET_FILL);
	glDisable(GL_BLEND);
}



int clicked_x = -1;
int clicked_y = -1;

void mouse_clicked(int x, int y)
{
    clicked_x = x;
    clicked_y = y;
}

void motion(int x, int y)
{
	Quat4f motion_from, motion_to;
	float distance;
    
	dragged_rot = UNIT_QUAT4F;
	if (clicked_x == x && clicked_y == y){
		return;
	}
    
	distance = sqrt((x-clicked_x)*(x-clicked_x)+(y-clicked_y)*(y-clicked_y));
    
	motion_from.w = 0;
	motion_from.x = (float)clicked_x / W - 0.5;  /* normalized clicked_x */
	motion_from.y = 0.5 - (float)clicked_y / H;  /* normalized clicked_y */
	motion_from.z = 1;  /* or 1-distance^2 */
	normalize_quat(&motion_from);
    
	motion_to.w = 0;
	motion_to.x = (float)x / W - 0.5;
	motion_to.y = 0.5 - (float)y / H;
	motion_to.z = 1;  /* or 1-distance^2 */
	normalize_quat(&motion_to);
    
	get_rot_quat_from_quats(&motion_to, &motion_from,
							atan(distance),  /* or acos(_to・_from) */
							&dragged_rot);
	normalize_quat(&dragged_rot);  /* supress error */
    
}

void mouse_released(int x, int y)
{
    if (x == clicked_x && y == clicked_y){  /* no dragging */
        return;
    }
    if (dragged_rot.w == 1.0){ /* maybe motion() was skipped */
        motion(x, y);
    }
    
    /* take dragged rotation in */
    mul_quat(&current_rot, &dragged_rot, &current_rot);
    
    /* set angular velocity of invert rotation */
    spring_rot = dragged_rot;
    scale_quat_angle(&spring_rot, -0.2);
    /* reset to no drag rotation */
    dragged_rot = UNIT_QUAT4F;
}

void resize(int w, int h)
{
    //	glViewport(0, 0, w, h);
    //	glMatrixMode(GL_PROJECTION);
    //	glLoadIdentity();
    //	glFrustum(-1, 1, -1, 1, 2, 10);
    //	glTranslated(0, 0, -5);  /* moving viewpoint backward */
    //	glMatrixMode(GL_MODELVIEW);
    
	W = w;
	H = h;
}

void timeout()
{
	mul_quat(&current_rot, &spring_rot, &current_rot);
	damp_quat_angle(&spring_rot, 0.01);
    //	if (spring_rot.x != 0 || spring_rot.y != 0 || spring_rot.z){
    //		glutTimerFunc(30, timeout, 0);
    //	}
    //	glutPostRedisplay();
}


//----------------------------------------------------------------------------------------------------------

@implementation GLViewController
- (void)drawView:(UIView *)theView
{
    float lightpos[] = {3.0f, 4.0f, 5.0f, 1.0f};
    
    glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
    
    //    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    if(1){
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        
        //lighting
        float lightpos[] = {3.0f, 4.0f, 5.0f, 1.0f};
        glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
        glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);
        
        //        [Ground draw];
        glDisable(GL_LIGHTING);
        
        //	ground_z -= 0.1;
        //	if (ground_z < -0.99) ground_z = 1.0;
    }
    if(1){
        draw();
    }
    if(1){
        //tmp, just for executed by timer
        [P2Object update];
        [Ground update];
        timeout();
    }

}

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
	UITouch* touch=[touches anyObject];
	CGPoint pt = [touch locationInView:self.view];
    mouse_clicked(pt.x, pt.y);
}

- (void)touchesMoved:(NSSet*)touches withEvent:(UIEvent*)event
{
	UITouch* touch=[touches anyObject];
	CGPoint pt = [touch locationInView:self.view];
    motion(pt.x, pt.y);
}

- (void)touchesEnded:(NSSet*)touches withEvent:(UIEvent*)event
{
	UITouch* touch=[touches anyObject];
	CGPoint pt = [touch locationInView:self.view];
    mouse_released(pt.x, pt.y);
}

- (void)touchesCancelled:(NSSet*)touches withEvent:(UIEvent*)event
{
}

-(void)setupView:(GLView*)view
{
	const GLfloat zNear = 0.01, zFar = 1000.0, fieldOfView = 45.0; 
	GLfloat size; 
	glEnable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION); 
	size = zNear * tanf(DEGREES_TO_RADIANS(fieldOfView) / 2.0); 
	CGRect rect = view.bounds; 
	glFrustumf(-size, size, -size / (rect.size.width / rect.size.height), size / 
			   (rect.size.width / rect.size.height), zNear, zFar); 
	glViewport(0, 0, rect.size.width, rect.size.height);  
	glMatrixMode(GL_MODELVIEW);

	//ynomura added
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	glFrustumf(-1, 1, -1, 1, 1.5f, 10);
    glTranslatef(0, 0, -3);
    glViewport(0, (rect.size.height - rect.size.width)/2, rect.size.width, rect.size.width);
	glMatrixMode(GL_MODELVIEW);
    
	glLoadIdentity();
}
- (void)dealloc 
{
    [super dealloc];
}
@end
