//
//  P2Object.m
//  TestApp3
//
//  Created by Yuji Nomura on 14/03/25.
//
//

#import "P2Object.h"
#import <OpenGLES/ES1/gl.h>
#import <OpenGLES/ES1/glext.h>
#include "MyQuat.h"

bool useTexture = true;
bool drawingShadow = false;

@implementation P2Object

/* ---------------- About VAO/VBO ---------------- */
#define NUM_OF_VERTEX_ARRAYS 6  /* [0] is reserved for testing */
const GLenum sphere_index_type = GL_UNSIGNED_SHORT;

enum Part {BODY, FLIPPER, FOOT, EYE, BEAK};

typedef struct {
	GLfloat vertex[3];
    GLfloat normal[3];
	GLfloat texCoord[2];
	GLfloat color[4];
} myVertexAttrib;

GLuint vaos[NUM_OF_VERTEX_ARRAYS];  /* an array of vertex array object IDs */
GLuint vbos[NUM_OF_VERTEX_ARRAYS];  /* vertex buffer object IDs correspond to VAOs */
GLuint eaos[NUM_OF_VERTEX_ARRAYS];  /* VBOs for element arrays correspond to VAOs */

int easizes[NUM_OF_VERTEX_ARRAYS];  /* an array of size of element arrays */
enum Part parts[NUM_OF_VERTEX_ARRAYS]; /* an array of part type */
int granularities[NUM_OF_VERTEX_ARRAYS];  /* an array of the granularity */
myVertexAttrib *vattrs[NUM_OF_VERTEX_ARRAYS]; /* an array of attribute array pointers */


#ifdef __GNUC__
#define myVertexAttribPointer(MEMBER) (void *)__builtin_offsetof(myVertexAttrib, MEMBER)
#else
myVertexAttrib dummyVertexAttrib;
#define myVertexAttribPointer(MEMBER) (void *)((int)&dummyVertexAttrib.MEMBER - (int)&dummyVertexAttrib)
#endif

void addIndexedOne8thSurface(int xsign, int ysign, int zsign, int granularity, myVertexAttrib *vaa, int vaoffset,  GLushort *indexp, int ioffset)
{
	Quat4f xunit = {0, xsign, 0, 0};
	Quat4f yunit = {0, 0, ysign, 0};
	Quat4f zunit = {0, 0, 0, zsign};
	Quat4f z1x, z1y, z2x, z2y;  /* start/end point on z1, z2 */
	Quat4f tmp;
	int a, b, c, d;  /* vertex indices of a parallelogram */
	int z, t;
	int vn = vaoffset;  /* number of vertices */
	int in = ioffset;  /* number of indices */
	int z1xi, z1yi, z2xi, z2yi;  /* vertex index of z1x, z1y, z2x, z2y */

	if (xsign * ysign * zsign == -1){
		/* swap some 2 of 3 vector */
		yunit.z = zsign; yunit.y = 0;
		zunit.y = ysign; zunit.z = 0;
	}

	z1xi = z1yi = vn;
	tmp = zunit;
	vaa[vn].vertex[0] = tmp.x;
	vaa[vn].vertex[1] = tmp.y;
	vaa[vn].vertex[2] = tmp.z;
	vaa[vn].normal[0] = tmp.x;
	vaa[vn].normal[1] = tmp.y;
	vaa[vn].normal[2] = tmp.z;
    /* counterclockwise from z=-1 (from back, in right-handed coordinate system) */
	vaa[vn].texCoord[0] = (atan2(tmp.x, tmp.z) / M_PI + 1) / 2;
    /* atan2(-0,-1) correction */
    if (vaa[vn].texCoord[0] == 1.0 && xsign == -1)
        vaa[vn].texCoord[0] = 0.0;
	vaa[vn].texCoord[1] = (tmp.y + 1) / 2;
	vn++;

	for(z = 1; z <= granularity; z++){
		slerp_quat(&zunit, &xunit, (float)z/granularity, &z2x);
		slerp_quat(&zunit, &yunit, (float)z/granularity, &z2y);
        
		z2xi = vn;
		z2yi = vn + z;
		for(t = 0; t <= z; t++){
			slerp_quat(&z2x, &z2y, (float)t/z, &tmp);
			vaa[vn].vertex[0] = tmp.x;
			vaa[vn].vertex[1] = tmp.y;
			vaa[vn].vertex[2] = tmp.z;
			vaa[vn].normal[0] = tmp.x;
			vaa[vn].normal[1] = tmp.y;
			vaa[vn].normal[2] = tmp.z;
			vaa[vn].texCoord[0] = (atan2(tmp.x, tmp.z) / M_PI + 1) / 2; /* counterclockwise from z=-1 */
            if (vaa[vn].texCoord[0] == 1.0 && xsign == -1) /* atan2(-0,-1) correction */
                vaa[vn].texCoord[0] = 0.0;
			vaa[vn].texCoord[1] = (tmp.y + 1) / 2;
			vn++;
		}
	
		/* triangle order by t:
		        a     c
		   z1x   ************  z1y
			    **  1 **  2 **
		       *  *  *  *  *  *
		      *t=0 **  1 **  2 *
		z2x  ********************  z2y
		            b     d    	       */

		/* leftmost(t=0) triangle */
		b = z2xi;
		c = z1xi;
		d = z2xi + 1;
		indexp[in++] = c;
		indexp[in++] = b;
		indexp[in++] = d;
		
		for(t = 1; t < z; t++){
			a = z1xi + t - 1;
			b = z2xi + t;
			c = z1xi + t;
			d = z2xi + t + 1;

			indexp[in++] = a;
			indexp[in++] = b;
			indexp[in++] = c;

			indexp[in++] = c;
			indexp[in++] = b;
			indexp[in++] = d;
		}

		z1xi = z2xi;
		z1yi = z2yi;
	}
}

void genPartsVertexArray(int va_No, enum Part part, int granularity)
{
	myVertexAttrib *attrs;
	int vn_per_one8th = (granularity+1)*(granularity+2)/2;
	GLushort *indices;
	int in_per_one8th = 3 * granularity*granularity;
	
	int i;
	float vx,vy,vz;
	float nx,ny,nz;
	float r,g,b,a;
    
	float dydx, dydz, d;  //for normal calculation
    
	attrs = malloc(sizeof(myVertexAttrib) * vn_per_one8th * 8);
	indices = malloc(sizeof(GLushort) * in_per_one8th * 8);
    if (!attrs || !indices) {
        NSLog(@"malloc() failed in genPartsVertexArray()");
        if (attrs) free(attrs);
        if (indices) free(indices);
        return;
    }

	addIndexedOne8thSurface( 1, 1, 1, granularity, attrs, vn_per_one8th * 0, indices, in_per_one8th * 0);
	addIndexedOne8thSurface( 1, 1,-1, granularity, attrs, vn_per_one8th * 1, indices, in_per_one8th * 1);
	addIndexedOne8thSurface( 1,-1, 1, granularity, attrs, vn_per_one8th * 2, indices, in_per_one8th * 2);
	addIndexedOne8thSurface( 1,-1,-1, granularity, attrs, vn_per_one8th * 3, indices, in_per_one8th * 3);
	addIndexedOne8thSurface(-1, 1, 1, granularity, attrs, vn_per_one8th * 4, indices, in_per_one8th * 4);
	addIndexedOne8thSurface(-1, 1,-1, granularity, attrs, vn_per_one8th * 5, indices, in_per_one8th * 5);
	addIndexedOne8thSurface(-1,-1, 1, granularity, attrs, vn_per_one8th * 6, indices, in_per_one8th * 6);
	addIndexedOne8thSurface(-1,-1,-1, granularity, attrs, vn_per_one8th * 7, indices, in_per_one8th * 7);


	/* modify vertex coords and vertex colors */
	for(i=0; i<vn_per_one8th * 8; i++){
		vx = attrs[i].vertex[0];
		vy = attrs[i].vertex[1];
		vz = attrs[i].vertex[2];

		nx = attrs[i].normal[0];
		ny = attrs[i].normal[1];
		nz = attrs[i].normal[2];

		switch(part){
		case BODY:
			/* modify vertices */
			if (vy < 0.2f) vy = 0.2 + (vy-0.2) * 0.8;
			else vy = 0.2 + (vy-0.2) * 0.9;

			/* update normal*/
			if (ny < 0.2f){
				dydx = -0.8 * nx / ny;
				dydz = -0.8 * nz / ny;
			}else{
				dydx = -0.9 * nx / ny;
				dydz = -0.9 * nz / ny;
			}

			if (ny < 0.0f){
				nx = dydx;
				ny = -1;
				nz = dydz;
			}
			else if (ny > 0.0f){
				nx = -dydx;
				ny = 1;
				nz = -dydz;
			}

			/* update color */
			if (vx * 0.0 + vy * (-0.5) + vz * 1.0 > 0.84){
				r = 1; g = 1; b = 1;
			}else{
				r = 0; g = 0; b = 1;
			}
			a = 1;
			break;
		case FLIPPER:
			break;
		case FOOT:
			vy = vy * 0.2;
			/* scale here */
			vx *= 0.36;
			vy *= 0.36;
			vz *= 0.48;
			/* update normal*/
			dydx = -0.36/0.36 * nx / ny;
			dydz = -0.36/0.48 * nz / ny;
			if (ny < 0.0f){
				nx = dydx;
				ny = -1;
				nz = dydz;
			}
			else if (ny > 0.0f){
				nx = -dydx;
				ny = 1;
				nz = -dydz;
			}
			/* color */
			r = 1; g = 1; b = 0; a = 1;
			break;
		case EYE:
			/* scale here */
			vx *= 0.08;
			vy *= 0.08;
			vz *= 0.08;
			/* color */
			r = 0; g = 0; b = 0; a = 1;
			break;
		case BEAK:
			vy = vy * 0.3;
			/* scale here */
			vx *= 0.4;
			vy *= 0.4;
			vz *= 0.4;
			/* update normal*/
			dydx = -0.3 * nx / ny;
			dydz = -0.3 * nz / ny;
			if (ny < 0.0f){
				nx = dydx;
				ny = -1;
				nz = dydz;
			}
			else if (ny > 0.0f){
				nx = -dydx;
				ny = 1;
				nz = -dydz;
			}
			/* color */
			r = 1; g = 1; b = 0; a = 1;
			break;
		default:
			break;
		}

		/* normalize normal */
		d = sqrt(nx*nx + ny*ny + nz*nz);
		nx /= d;
		ny /= d;
		nz /= d;
		
		attrs[i].vertex[0] = vx;
		attrs[i].vertex[1] = vy;
		attrs[i].vertex[2] = vz;

		attrs[i].normal[0] = nx;
		attrs[i].normal[1] = ny;
		attrs[i].normal[2] = nz;

		attrs[i].color[0] = r;
		attrs[i].color[1] = g;
		attrs[i].color[2] = b;
		attrs[i].color[3] = a;
    }

	glBindVertexArrayOES(vaos[va_No]);

	glBindBuffer(GL_ARRAY_BUFFER, vbos[va_No]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(myVertexAttrib) * vn_per_one8th * 8, attrs, GL_STATIC_DRAW);
	glVertexPointer(3, GL_FLOAT, sizeof(myVertexAttrib), myVertexAttribPointer(vertex));

    glNormalPointer(GL_FLOAT, sizeof(myVertexAttrib), myVertexAttribPointer(normal));

	glTexCoordPointer(2, GL_FLOAT, sizeof(myVertexAttrib), myVertexAttribPointer(texCoord));
	glColorPointer(4, GL_FLOAT, sizeof(myVertexAttrib), myVertexAttribPointer(color));

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	
	glBindVertexArrayOES(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	/* store vertex index array in VBO (because VAO can't contain this) */
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, eaos[va_No]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * in_per_one8th * 8, indices, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	easizes[va_No] = in_per_one8th * 8;
    
	if (part == FLIPPER){
		vattrs[va_No] = attrs;
	}else{
	    free(attrs);
		vattrs[va_No] = NULL;
	}
    
	parts[va_No] = part;
	granularities[va_No] = granularity;
    
	free(indices);
}

void morphPartsArray(int va_No, int morph_param)
{
	myVertexAttrib *attrs;
	int granularity = granularities[va_No];
	enum Part part = parts[va_No];
	int vn_per_one8th = (granularity+1)*(granularity+2)/2;
	
	int i;
	float vx,vy,vz;
	float nx,ny,nz;

	float d;

	attrs = malloc(sizeof(myVertexAttrib) * vn_per_one8th * 8);
    if (!attrs) {
        NSLog(@"malloc() failed in morphPartsArray()");
        return;
    }
    memcpy(attrs, vattrs[va_No], sizeof(myVertexAttrib) * vn_per_one8th * 8);

	for(i=0; i<vn_per_one8th * 8; i++){
		vx = attrs[i].vertex[0];
		vy = attrs[i].vertex[1];
		vz = attrs[i].vertex[2];

		nx = attrs[i].normal[0];
		ny = attrs[i].normal[1];
		nz = attrs[i].normal[2];

		switch(part){
		case FLIPPER:
			vy = vy * 0.1 + 0.12 * morph_param * (vx*vx * (1 + vz*vz));

            /* scale here */
            vx *= 0.7;
            vy *= 0.6;
            vz *= 0.5;

			if (ny != 0){
				float dydx, dydz;  /* dy/dx, dy/dz */

				/* let Y = a y + b vx^2 (1 + vz^2)
				   dY/dx = a dy/dx + 2b (1 + vz^2) x
				         = -a x/y + 2b (1 + vz^2) x   #dy/dx = -x/y where x^2+y^2=1
				   dY/dz = -a x/y + 2b x^2 z
				*/
				dydx = -0.1 * nx / ny + 2 * 0.12 * morph_param * nx * (1 + nz*nz);
				dydz = -0.1 * nz / ny + 2 * 0.12 * morph_param * nx*nx * nz;

				/* vertex y>0 -> normal y>0
				   vertex y<0 -> normal y<0
				   as downward face keeps downward after morphing
				   therefore normal is either (-y',1) or (y', 1)
				*/
				if (ny > 0){
					nx = -dydx;
					ny = 1;
					nz = -dydz;
				}else{
					nx = dydx;
					ny = -1;
					nz = dydz;
				}
            }                    

			/* fall through */
		default:
			break;
		}

		/* normalize normal */
		d = sqrt(nx*nx + ny*ny + nz*nz);
		nx /= d;
		ny /= d;
		nz /= d;

		attrs[i].vertex[0] = vx;
		attrs[i].vertex[1] = vy;
		attrs[i].vertex[2] = vz;

		attrs[i].normal[0] = nx;
		attrs[i].normal[1] = ny;
		attrs[i].normal[2] = nz;
    }
    
	glBindVertexArrayOES(vaos[va_No]);
    
	glBindBuffer(GL_ARRAY_BUFFER, vbos[va_No]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(myVertexAttrib) * vn_per_one8th * 8, attrs, GL_STATIC_DRAW);
	glVertexPointer(3, GL_FLOAT, sizeof(myVertexAttrib), myVertexAttribPointer(vertex));

    glNormalPointer(GL_FLOAT, sizeof(myVertexAttrib), myVertexAttribPointer(normal));

	glTexCoordPointer(2, GL_FLOAT, sizeof(myVertexAttrib), myVertexAttribPointer(texCoord));
	glColorPointer(4, GL_FLOAT, sizeof(myVertexAttrib), myVertexAttribPointer(color));

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	
	glBindVertexArrayOES(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

    free(attrs);
}

/* ---------------- End of somethings about VAO/VBO ---------------- */

/* ---------------- Scene graph scheme and utilities ---------------- */
enum NodeType {TYPE_NODE, TYPE_SCALE, TYPE_TRANSLATE, TYPE_ROTATE, TYPE_VERTEXARRAY, TYPE_INDEXED_VERTEXARRAY, TYPE_VERTEXBUFFER};

typedef struct _Node {
    int type;
    struct _Node *child[7];
} Node;

typedef struct {
    Node node;
    GLfloat	scale_x, scale_y, scale_z;
} ScaleNode;

typedef struct {
    Node node;
    GLfloat	tr_x, tr_y, tr_z;
} TranslateNode;

typedef struct {
    Node node;
    GLfloat	angle, x, y, z;
} RotateNode;

typedef struct {
    Node node;
    GLuint va;
} VertexArrayNode;

typedef struct {
    Node node;
    GLuint va;
    GLuint index_buf;
    int index_size;
    GLuint color;
    GLuint use_texture;
} IndexedVertexArrayNode;

typedef struct {
    Node node;
    GLuint va;
    GLuint index_buf;
    int index_size;
    GLuint vb;  /* can be generated and deleted at every morphing? */
    const myVertexAttrib *attrs;  /* original, copied at morphing */
} VertexBufferNode;

void drawScene(Node *root)
{
    ScaleNode *s;
    TranslateNode *t;
    RotateNode *r;
    IndexedVertexArrayNode *iva;
    int i;

    if (root == NULL) return;

	glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

    switch(root->type){
	case TYPE_SCALE:
		s = (ScaleNode *)root;
		glScalef(s->scale_x, s->scale_y, s->scale_z);
		break;
	case TYPE_TRANSLATE:
		t = (TranslateNode *)root;
		glTranslatef(t->tr_x, t->tr_y, t->tr_z);
		break;
	case TYPE_ROTATE:
		r = (RotateNode *)root;
		glRotatef(r->angle, r->x, r->y, r->z);
		break;
	case TYPE_VERTEXARRAY:
		break;
	case TYPE_INDEXED_VERTEXARRAY:
		iva = (IndexedVertexArrayNode *)root;
		glBindVertexArrayOES(iva->va);

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);

        if (drawingShadow) {
             /* no material/texture operation */
        }
		else if (iva->use_texture) {
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glBindTexture(GL_TEXTURE_2D, iva->use_texture);
			glEnable(GL_TEXTURE_2D);
			//Texture will be multiplied by color fragment(as TEXTURE_ENV_MODE is MODULATE), so
			//Material properties should be set.
			//Instead of using glMaterialfv(), enable GL_COLOR_MATERIAL and use glColor4 here.
			glEnable(GL_COLOR_MATERIAL);
			if (iva->color)
				glColor4ub(iva->color >> 24, iva->color >> 16, iva->color >> 8, iva->color);
			else
				glColor4f(1, 1, 1, 1);
			glDisable(GL_COLOR_MATERIAL);  //color material cancels texture
		} else {
			//current color or color material
			if (iva->color) {
				//Instead of using glMaterialfv(), enable GL_COLOR_MATERIAL and using glColor4 here
				glEnable(GL_COLOR_MATERIAL);
				glColor4ub(iva->color >> 24, iva->color >> 16, iva->color >> 8, iva->color);
				glDisable(GL_COLOR_MATERIAL);  //color material cancels current color
				glDisableClientState(GL_COLOR_ARRAY);
			} else {
				//use color array
				glEnableClientState(GL_COLOR_ARRAY);
				glEnable(GL_COLOR_MATERIAL);
			}
		}

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, iva->index_buf);
		glDrawElements(GL_TRIANGLES, iva->index_size, GL_UNSIGNED_SHORT, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

		if (iva->use_texture) {
			glDisable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, 0);
		}
		glDisable(GL_COLOR_MATERIAL);
            
		glBindVertexArrayOES(0);
		break;
	case TYPE_VERTEXBUFFER:
		break;
	default:
		;
    }

    for(i=0; i<sizeof(root->child)/sizeof(root->child[0]); i++){
        if (root->child[i] != NULL){
            drawScene(root->child[i]);
        }
    }

    glPopMatrix();
}
    
/* ---------------- End of scene graph things ---------------- */

/* ---------------- Start of application code ---------------- */

Node scene_root = {TYPE_NODE};
GLuint texture;

void createScene(void)
{
	 /*
           5          6       (eyes)
                7             (beak)
           1    0     2       (flipper, body, flipper)
             3     4          (feet)
      */

	static TranslateNode tn0 = {{TYPE_TRANSLATE}};
	static TranslateNode tn1 = {{TYPE_TRANSLATE}};
	static TranslateNode tn2 = {{TYPE_TRANSLATE}};
	static TranslateNode tn3 = {{TYPE_TRANSLATE}};
	static TranslateNode tn4 = {{TYPE_TRANSLATE}};
	static TranslateNode tn5 = {{TYPE_TRANSLATE}};
	static TranslateNode tn6 = {{TYPE_TRANSLATE}};
	static TranslateNode tn7 = {{TYPE_TRANSLATE}};
	static RotateNode rn0 = {{TYPE_ROTATE}};
	static RotateNode rn1 = {{TYPE_ROTATE}};
	static RotateNode rn2 = {{TYPE_ROTATE}};
	static RotateNode rn3 = {{TYPE_ROTATE}};
	static RotateNode rn4 = {{TYPE_ROTATE}};
	static RotateNode rn7 = {{TYPE_ROTATE}};

	static IndexedVertexArrayNode ivn0 = {{TYPE_INDEXED_VERTEXARRAY}};
	static IndexedVertexArrayNode ivn1 = {{TYPE_INDEXED_VERTEXARRAY}};
	static IndexedVertexArrayNode ivn2 = {{TYPE_INDEXED_VERTEXARRAY}};
	static IndexedVertexArrayNode ivn3 = {{TYPE_INDEXED_VERTEXARRAY}};
	static IndexedVertexArrayNode ivn4 = {{TYPE_INDEXED_VERTEXARRAY}};
	static IndexedVertexArrayNode ivn5 = {{TYPE_INDEXED_VERTEXARRAY}};
	static IndexedVertexArrayNode ivn6 = {{TYPE_INDEXED_VERTEXARRAY}};
	static IndexedVertexArrayNode ivn7 = {{TYPE_INDEXED_VERTEXARRAY}};
    
    scene_root.child[0] = (Node *)&tn0;
    tn0.node.child[0] = (Node *)&rn0;  tn1.tr_x = tn1.tr_y = tn1.tr_z = 0;
    rn0.node.child[0] = (Node *)&ivn0; rn0.x = 0; rn0.y = 0; rn0.z = 1; rn0.angle = 30;
    ivn0.node.child[0] = (Node *)&tn1;
    ivn0.node.child[1] = (Node *)&tn2;
    ivn0.node.child[2] = (Node *)&rn3;
    ivn0.node.child[3] = (Node *)&rn4;
    ivn0.node.child[4] = (Node *)&tn5;
    ivn0.node.child[5] = (Node *)&tn6;
    ivn0.node.child[6] = (Node *)&tn7;
    
    tn1.node.child[0] = (Node *)&rn1;  tn1.tr_x =  0.8;  tn1.tr_y =  0.2;
    tn2.node.child[0] = (Node *)&rn2;  tn2.tr_x = -0.8;  tn2.tr_y =  0.2;
    rn3.node.child[0] = (Node *)&tn3; rn3.x = 1; rn3.y = 0; rn3.z = 0; rn3.angle = 30;
    rn4.node.child[0] = (Node *)&tn4; rn4.x = 1; rn4.y = 0; rn4.z = 0; rn4.angle = -30;
    tn3.node.child[0] = (Node *)&ivn3;  tn3.tr_x =  0.5;  tn3.tr_y = -0.7;  tn3.tr_z = 0.2;
    tn4.node.child[0] = (Node *)&ivn4;  tn4.tr_x = -0.5;  tn4.tr_y = -0.7;  tn4.tr_z = 0.2;
    tn5.node.child[0] = (Node *)&ivn5;  tn5.tr_x = -0.25;  tn5.tr_y =  0.45;  tn5.tr_z = 0.8;
    tn6.node.child[0] = (Node *)&ivn6;  tn6.tr_x =  0.25;  tn6.tr_y =  0.45;  tn6.tr_z = 0.8;
    tn7.node.child[0] = (Node *)&rn7;  tn7.tr_x =  0.0;  tn7.tr_y =  0.2;  tn7.tr_z = 0.7;
    
    rn1.node.child[0] = (Node *)&ivn1; rn1.x = 0; rn1.y = 0; rn1.z = 1; rn1.angle = -20;
    rn2.node.child[0] = (Node *)&ivn2; rn2.x = 0; rn2.y = 0; rn2.z = 1; rn2.angle = 20;
    rn7.node.child[0] = (Node *)&ivn7; rn7.x = 1; rn7.y = 0; rn7.z = 0; rn7.angle = -15;
    
    ivn0.va = vaos[1]; ivn0.index_buf = eaos[1]; ivn0.index_size = easizes[1]; ivn0.use_texture = /* texture */ useTexture;
    ivn1.va = vaos[2]; ivn1.index_buf = eaos[2]; ivn1.index_size = easizes[2]; ivn1.color = 0x0000ffff;
    ivn2.va = vaos[2]; ivn2.index_buf = eaos[2]; ivn2.index_size = easizes[2]; ivn2.color = 0x0000ffff;
    ivn3.va = vaos[3]; ivn3.index_buf = eaos[3]; ivn3.index_size = easizes[3]; ivn3.color = 0xffff00ff;
    ivn4.va = vaos[3]; ivn4.index_buf = eaos[3]; ivn4.index_size = easizes[3]; ivn4.color = 0xffff00ff;
    ivn5.va = vaos[4]; ivn5.index_buf = eaos[4]; ivn5.index_size = easizes[4]; ivn5.color = 0x000000ff;
    ivn6.va = vaos[4]; ivn6.index_buf = eaos[4]; ivn6.index_size = easizes[4]; ivn6.color = 0x000000ff;
    ivn7.va = vaos[5]; ivn7.index_buf = eaos[5]; ivn7.index_size = easizes[5]; ivn7.color = 0xffff00ff;
}

+ (void)update
{
	static int time = 0;
	int round_trip_value;

	/* for flipper motion */
	int morph_param;

	/* for foot motion */
	int foot_angle;
	TranslateNode *tn0 = (TranslateNode *)scene_root.child[0];
	RotateNode *rn0 = (RotateNode *)scene_root.child[0]->child[0];
	RotateNode *rn3 = (RotateNode *)scene_root.child[0]->child[0]->child[0]->child[2];
	RotateNode *rn4 = (RotateNode *)scene_root.child[0]->child[0]->child[0]->child[3];

	time++;
	
	/* flipper motion */
	if (time % 80 >= 40) {
		round_trip_value = (time + 2) % 20;
		if (round_trip_value > 10)
			round_trip_value = 20 - round_trip_value;
		morph_param = round_trip_value - 5;
	} else {
		morph_param = -3;
	}
	morphPartsArray(2, morph_param);

	/* foot motion */
	round_trip_value = time % 40;
	if (round_trip_value > 20)
		round_trip_value = 40 - round_trip_value;

	foot_angle = round_trip_value * 2 - 20;
	rn3->angle = foot_angle;
	rn4->angle = -foot_angle;
	rn0->angle = -foot_angle * 0.2;
	tn0->tr_y = foot_angle * foot_angle * 0.0001;
}

//#define TEX_WIDTH  940  /* =(300*3.14) */
//#define TEX_HEIGHT 300
#define TEX_WIDTH  512
#define TEX_HEIGHT 256
#define dst2(x,y) ((x)*(x)+(y)*(y))
#define is_near(x, a, b) ((x) >= (a) - (b) && (x) <= (a) + (b))

void initTexture()
{
	unsigned char *tex = malloc(3 * TEX_WIDTH * TEX_HEIGHT);
    int x,y;
	
	/* texture image for body */
	for (y=0;y<TEX_HEIGHT;y++) {
		for (x=0;x<TEX_WIDTH;x++) {
			unsigned char red = 0;
			unsigned char green = 0;
			unsigned char blue = 255;
            
//			if (dst2(x - TEX_WIDTH /2, y - 100) < 10000){
            if (dst2((x - TEX_WIDTH /2) *3/2, y - 75) < 7000){
				red = green = 255;
			}
            
//			if (y < 40){
			if (y < 34){
				red = green = 255;
			}
            
            tex[(x + y*TEX_WIDTH)*3 + 0] = red;
            tex[(x + y*TEX_WIDTH)*3 + 1] = green;
            tex[(x + y*TEX_WIDTH)*3 + 2] = blue;
		}
	}
    
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
    
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, TEX_WIDTH, TEX_HEIGHT, 0, GL_RGB, GL_UNSIGNED_BYTE, tex);
	free(tex);
	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);  //necessary for non-power-of-two textures
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
}

+(void)draw
{
    drawScene(&scene_root);
}

+(void)drawShadow
{
    drawingShadow = true;
    drawScene(&scene_root);
    drawingShadow = false;
}

+(void)initialize
{
	glGenVertexArraysOES(NUM_OF_VERTEX_ARRAYS, vaos);
	glGenBuffers(NUM_OF_VERTEX_ARRAYS, vbos);
	glGenBuffers(NUM_OF_VERTEX_ARRAYS, eaos);
    
    genPartsVertexArray(1, BODY, 20);
    genPartsVertexArray(2, FLIPPER, 10);
	genPartsVertexArray(3, FOOT, 5);
	genPartsVertexArray(4, EYE, 5);
	genPartsVertexArray(5, BEAK, 5);
	
	morphPartsArray(2, -3);

    initTexture();

    createScene();
}

@end
